# Copyright 2024 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from typing import Set

from rp2.abstract_country import AbstractCountry
from rp2.rp2_error import RP2ValueError
from rp2.rp2_main import rp2_main


# Generic country class
class Generic(AbstractCountry):
    def __init__(self) -> None:
        currency_iso_code = os.environ.get("CURRENCY_CODE")
        if not currency_iso_code:
            raise RP2ValueError("CURRENCY_CODE environment variable not found: it is required to use the generic country plugin")
        long_term_capital_gain_period = os.environ.get("LONG_TERM_CAPITAL_GAINS")
        if not long_term_capital_gain_period:
            raise RP2ValueError("LONG_TERM_CAPITAL_GAINS environment variable not found: it is required to use the generic country plugin")
        try:
            self.__long_term_capital_gain_period = int(long_term_capital_gain_period)
        except (ValueError, TypeError) as exc:
            raise RP2ValueError(
                f"LONG_TERM_CAPITAL_GAINS environment variable has value " f"'{long_term_capital_gain_period}', which is not convertible to integer"
            ) from exc
        if self.__long_term_capital_gain_period < 0:
            raise RP2ValueError(
                f"LONG_TERM_CAPITAL_GAINS environment variable has negative value " f"'{long_term_capital_gain_period}': it should be >= 0 instead"
            )
        super().__init__("generic", currency_iso_code)

    # Measured in days
    def get_long_term_capital_gain_period(self) -> int:
        return self.__long_term_capital_gain_period

    # Default accounting method to use if the user doesn't specify one on the command line
    def get_default_accounting_method(self) -> str:
        return "fifo"

    # Set of accounting methods accepted in the country
    def get_accounting_methods(self) -> Set[str]:
        return {"fifo", "lifo", "hifo"}

    # Default set of generators to use if the user doesn't specify them on the command line
    def get_report_generators(self) -> Set[str]:
        return {
            "open_positions",
            "rp2_full_report",
        }

    # Default language to use at report generation if the user doesn't specify it on the command line (in ISO 639-1 format)
    def get_default_generation_language(self) -> str:
        return "en"


# US-specific entry point
def rp2_entry() -> None:
    rp2_main(Generic())



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
from heapq import heappop, heappush
from typing import Dict, List, NamedTuple, Optional, Tuple

from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError


class AcquiredLotAndAmount(NamedTuple):
    acquired_lot: InTransaction
    amount: RP2Decimal


class AcquiredLotCandidatesOrder(Enum):
    OLDER_TO_NEWER: str = "older_to_newer"
    NEWER_TO_OLDER: str = "newer_to_older"


class AcquiredLotHeapSortKey(NamedTuple):
    spot_price: RP2Decimal
    timestamp: float
    internal_id_int: int


class AbstractAccountingMethodIterator:
    def __next__(self) -> InTransaction:
        raise NotImplementedError("abstract function")


class ListAccountingMethodIterator(AbstractAccountingMethodIterator):
    def __init__(self, acquired_lot_list: List[InTransaction], from_index: int, to_index: int, order_type: AcquiredLotCandidatesOrder) -> None:
        self.__acquired_lot_list = acquired_lot_list
        self.__start_index = from_index if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else to_index
        self.__end_index = to_index if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else from_index
        self.__step = 1 if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else -1
        self.__index = self.__start_index
        self.__order_type = order_type

    def _check_index(self) -> bool:
        if self.__order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER:
            return self.__index <= self.__end_index
        return self.__index >= self.__end_index

    def __next__(self) -> InTransaction:
        result: Optional[InTransaction] = None
        while self._check_index():
            result = self.__acquired_lot_list[self.__index]
            self.__index += self.__step
            return result
        raise StopIteration(self)


class HeapAccountingMethodIterator(AbstractAccountingMethodIterator):
    def __init__(self, acquired_lot_heap: List[Tuple[AcquiredLotHeapSortKey, InTransaction]]) -> None:
        self.__acquired_lot_heap = acquired_lot_heap

    def __next__(self) -> InTransaction:
        while len(self.__acquired_lot_heap) > 0:
            _, result = heappop(self.__acquired_lot_heap)
            return result
        raise StopIteration(self)


class AbstractAcquiredLotCandidates:
    def __init__(
        self,
        accounting_method: "AbstractAccountingMethod",
        acquired_lot_list: List[InTransaction],
        acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal],
    ) -> None:
        self._accounting_method: AbstractAccountingMethod = accounting_method
        self.__acquired_lot_list = acquired_lot_list
        self.__acquired_lot_2_partial_amount = acquired_lot_2_partial_amount
        self.__to_index = 0
        self.__from_index = 0

    def set_from_index(self, from_index: int) -> None:
        self.__from_index = from_index

    def set_to_index(self, to_index: int) -> None:
        self.__to_index = to_index  # pylint: disable=unused-private-member

    @property
    def from_index(self) -> int:
        return self.__from_index

    @property
    def to_index(self) -> int:
        return self.__to_index

    @property
    def acquired_lot_list(self) -> List[InTransaction]:
        return self.__acquired_lot_list

    def has_partial_amount(self, acquired_lot: InTransaction) -> bool:
        return acquired_lot in self.__acquired_lot_2_partial_amount

    def get_partial_amount(self, acquired_lot: InTransaction) -> RP2Decimal:
        if not self.has_partial_amount(acquired_lot):
            raise RP2RuntimeError(f"Internal error: acquired lot has no partial amount: {acquired_lot}")
        return self.__acquired_lot_2_partial_amount[acquired_lot]

    def set_partial_amount(self, acquired_lot: InTransaction, amount: RP2Decimal) -> None:
        self.__acquired_lot_2_partial_amount[acquired_lot] = amount

    def clear_partial_amount(self, acquired_lot: InTransaction) -> None:
        self.set_partial_amount(acquired_lot, ZERO)

    def __iter__(self) -> AbstractAccountingMethodIterator:
        return self._accounting_method._create_accounting_method_iterator(self)


class ListAcquiredLotCandidates(AbstractAcquiredLotCandidates):
    pass


class HeapAcquiredLotCandidates(AbstractAcquiredLotCandidates):
    _accounting_method: "AbstractHeapAccountingMethod"

    def __init__(
        self,
        accounting_method: "AbstractAccountingMethod",
        acquired_lot_list: List[InTransaction],
        acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal],
    ) -> None:
        super().__init__(accounting_method, acquired_lot_list, acquired_lot_2_partial_amount)
        self.__acquired_lot_heap: List[Tuple[AcquiredLotHeapSortKey, InTransaction]] = []

    def set_to_index(self, to_index: int) -> None:
        # Control how far to advance the iterator, caller is responsible for updating
        for i in range(self.to_index, to_index + 1):
            lot = self.acquired_lot_list[i]
            self._accounting_method.add_selected_lot_to_heap(self.acquired_lot_heap, lot)
        super().set_to_index(to_index)

    @property
    def acquired_lot_heap(self) -> List[Tuple[AcquiredLotHeapSortKey, InTransaction]]:
        return self.__acquired_lot_heap


class AbstractAccountingMethod:
    def create_lot_candidates(
        self, acquired_lot_list: List[InTransaction], acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal]
    ) -> AbstractAcquiredLotCandidates:
        raise NotImplementedError("abstract")

    def seek_non_exhausted_acquired_lot(
        self,
        lot_candidates: AbstractAcquiredLotCandidates,
        taxable_event: Optional[AbstractTransaction],
        taxable_event_amount: RP2Decimal,
    ) -> Optional[AcquiredLotAndAmount]:
        raise NotImplementedError("Abstract function")

    def lot_candidates_order(self) -> AcquiredLotCandidatesOrder:
        raise NotImplementedError("Abstract function")

    @property
    def name(self) -> str:
        return f"{self.__class__.__module__.rsplit('.', 1)[1]}"

    def __repr__(self) -> str:
        return self.name

    def _create_accounting_method_iterator(self, lot_candidates: AbstractAcquiredLotCandidates) -> AbstractAccountingMethodIterator:
        raise NotImplementedError("Abstract function")


class AbstractListAccountingMethod(AbstractAccountingMethod):
    def create_lot_candidates(
        self, acquired_lot_list: List[InTransaction], acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal]
    ) -> ListAcquiredLotCandidates:
        return ListAcquiredLotCandidates(self, acquired_lot_list, acquired_lot_2_partial_amount)

    def lot_candidates_order(self) -> AcquiredLotCandidatesOrder:
        raise NotImplementedError("Abstract function")

    def _create_accounting_method_iterator(self, lot_candidates: AbstractAcquiredLotCandidates) -> ListAccountingMethodIterator:
        return ListAccountingMethodIterator(lot_candidates.acquired_lot_list, lot_candidates.from_index, lot_candidates.to_index, self.lot_candidates_order())


class AbstractHeapAccountingMethod(AbstractAccountingMethod):
    def create_lot_candidates(
        self, acquired_lot_list: List[InTransaction], acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal]
    ) -> HeapAcquiredLotCandidates:
        return HeapAcquiredLotCandidates(self, acquired_lot_list, acquired_lot_2_partial_amount)

    def add_selected_lot_to_heap(self, heap: List[Tuple[AcquiredLotHeapSortKey, InTransaction]], lot: InTransaction) -> None:
        heap_item = (self.heap_key(lot), lot)
        heappush(heap, heap_item)

    def heap_key(self, lot: InTransaction) -> AcquiredLotHeapSortKey:
        raise NotImplementedError("Abstract function")

    def _create_accounting_method_iterator(self, lot_candidates: AbstractAcquiredLotCandidates) -> HeapAccountingMethodIterator:
        if not isinstance(lot_candidates, HeapAcquiredLotCandidates):
            raise RP2TypeError(f"Internal error: lot_candidates is not of type HeapAcquiredLotCandidates, but of type {type(lot_candidates)}")
        return HeapAccountingMethodIterator(lot_candidates.acquired_lot_heap)



# Copyright 2024 qwhelan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import date

from rp2.balance import BalanceSet
from rp2.configuration import Configuration
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.out_transaction import OutTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2ValueError
from rp2.transaction_set import TransactionSet


class TestBalanceSet(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestBalanceSet._configuration = Configuration("./config/test_data.ini", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_easy_negative_case(self) -> None:
        """
        Check that an exception is raised in the case where sum(OUT) > sum(IN)
        """
        asset = "B1"
        end_date = date(2024, 1, 1)
        in_transaction_set: TransactionSet = TransactionSet(self._configuration, "IN", asset)
        out_transaction_set: TransactionSet = TransactionSet(self._configuration, "OUT", asset)
        intra_transaction_set: TransactionSet = TransactionSet(self._configuration, "INTRA", asset)

        transaction1: InTransaction = InTransaction(
            self._configuration,
            "1/8/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "BUY",
            RP2Decimal("1000"),
            RP2Decimal("3.0002"),
            fiat_fee=RP2Decimal("20"),
            fiat_in_no_fee=RP2Decimal("3000.2"),
            fiat_in_with_fee=RP2Decimal("3020.2"),
            row=30,
        )
        in_transaction_set.add_entry(transaction1)

        transaction2: OutTransaction = OutTransaction(
            self._configuration,
            "1/9/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "SELL",
            RP2Decimal("1000"),
            RP2Decimal("4.0000"),
            crypto_fee=RP2Decimal("0.0002"),
            fiat_out_no_fee=RP2Decimal("4000.0"),
            row=31,
        )
        out_transaction_set.add_entry(transaction2)

        input_data = InputData(asset, in_transaction_set, out_transaction_set, intra_transaction_set)

        with self.assertRaisesRegex(
            RP2ValueError, r'B1 balance of account "Coinbase" \(holder "Alice"\) went negative \(-1.0000\) on the following transaction: .*'
        ):
            BalanceSet(self._configuration, input_data, end_date)

    def test_hard_negative_case(self) -> None:
        """
        Check that an exception is raised in the case where sum(OUT) > sum(IN) only briefly
        """
        asset = "B1"
        end_date = date(2024, 1, 1)
        in_transaction_set: TransactionSet = TransactionSet(self._configuration, "IN", asset)
        out_transaction_set: TransactionSet = TransactionSet(self._configuration, "OUT", asset)
        intra_transaction_set: TransactionSet = TransactionSet(self._configuration, "INTRA", asset)

        transaction1: InTransaction = InTransaction(
            self._configuration,
            "1/8/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "BUY",
            RP2Decimal("1000"),
            RP2Decimal("3.0002"),
            fiat_fee=RP2Decimal("20"),
            fiat_in_no_fee=RP2Decimal("3000.2"),
            fiat_in_with_fee=RP2Decimal("3020.2"),
            row=30,
        )
        in_transaction_set.add_entry(transaction1)

        transaction2: OutTransaction = OutTransaction(
            self._configuration,
            "1/9/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "SELL",
            RP2Decimal("1000"),
            RP2Decimal("4.0000"),
            crypto_fee=RP2Decimal("0.0002"),
            fiat_out_no_fee=RP2Decimal("6000.0"),
            row=31,
        )
        out_transaction_set.add_entry(transaction2)

        transaction3: InTransaction = InTransaction(
            self._configuration,
            "1/10/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "BUY",
            RP2Decimal("1000"),
            RP2Decimal("3.0002"),
            fiat_fee=RP2Decimal("20"),
            fiat_in_no_fee=RP2Decimal("3000.2"),
            fiat_in_with_fee=RP2Decimal("3020.2"),
            row=32,
        )
        in_transaction_set.add_entry(transaction3)

        transaction4: OutTransaction = OutTransaction(
            self._configuration,
            "1/11/2021 8:42:43.883 -04:00",
            asset,
            "Coinbase",
            "Alice",
            "SELL",
            RP2Decimal("1000"),
            RP2Decimal("2.0000"),
            crypto_fee=RP2Decimal("0.0002"),
            fiat_out_no_fee=RP2Decimal("2000.0"),
            row=33,
        )
        out_transaction_set.add_entry(transaction4)

        input_data = InputData(asset, in_transaction_set, out_transaction_set, intra_transaction_set)

        with self.assertRaisesRegex(
            RP2ValueError, r'B1 balance of account "Coinbase" \(holder "Alice"\) went negative \(-1.0000\) on the following transaction: .*'
        ):
            BalanceSet(self._configuration, input_data, end_date)


if __name__ == "__main__":
    unittest.main()
